/*
Suppose that DS is the type of some data structure.

Here are six patterns for functions that use a DS
data structure as either an input or an output value.

The first three are used all the time.
The fourth is rarely used (see GETADDRINFO(3) for an example).
The last two are why C++ has copy constructors.

DS * f(...);               // callee allocates DS, caller may or may not de-allocate

     f(const DS * x, ...); // caller allocates DS. DS is strictly an input

     f(DS * x, ...);       // caller allocates DS. DS may be input and/or output.

     f(DS ** x, ...);      // probably callee allocates DS and caller de-allocates it

DS   f(...);               // DS automatically allocated in caller's local space

     f(DS x, ...);         // DS automatically allocated in callee's local space


Below is an example of each possible use of a data structure by a function.
*/
#include <stdio.h>
#include <stdlib.h>


typedef struct DS {
   int a;
   double b;
   char c;
} DS;


DS * f1a()  /* callee allocates DS, caller may not de-allocate */
{
   static DS x;
   x.a = 1;
   return &x;
}


DS * f1b()  /* callee allocates DS, caller must de-allocate */
{
   DS * x = (DS *)malloc(sizeof(DS));
   x->a = 1;
   return x;
}


void f2(const DS * x)  /* caller allocates DS. DS is strictly an input */
{
   printf("in f2(), x->a = %d\n", x->a);
// x->a = -1;         /* this does not compile */
   ((DS *)x)->a = -1; /* this does compile, and breaks the const declaration */
   DS * y = x;        /* the previous line is equivalent to this and the next line */
   y->a = -1;
   x++;               /* notice that x is NOT constant (what x points at is) */
}


void f3(DS * x) /* caller allocates DS. DS may be input and/or output */
{
   printf("in f3(), x->a = %d\n", x->a);
   x->a = 3;
}


void f4(DS ** x)  /* callee allocates DS and caller de-allocates it */
{
   *x = (DS *)malloc(sizeof(DS));
   (*x)->a = 4;
}


DS f5()  /* DS automatically allocated in caller's local space */
{
   DS x;
   x.a = 5;
   return x;
}


void f6(DS x)  /* DS automatically allocated in callee's local space */
{
   printf("in f6(), x.a = %d\n", x.a);
}


int main()
{
   DS * ds1a_p = f1a();
   printf("in main(), ds1a_p->a = %d\n", ds1a_p->a);

   DS * ds1b_p = f1b();
   printf("in main(), ds1b_p->a = %d\n", ds1b_p->a);
   free(ds1b_p);

   DS ds2;    /* we could dynamically allocate ds2 */
   ds2.a = 2;
   f2( &ds2 );
   printf("in main(), ds2.a = %d\n", ds2.a);

   DS ds3;    /* we could dynamically allocate ds3 */
   ds3.a = -1;
   f3( &ds3 );
   printf("in main(), ds3.a = %d\n", ds3.a);

   DS * ds4_p;
   f4( &ds4_p );
   printf("in main(), ds4_p->a = %d\n", ds4_p->a);
   free( ds4_p );

   DS ds5 = f5(); /* return value from f5() is copied into ds5 */
   printf("in main(), ds5.a = %d\n", ds5.a);

   DS ds6;
   ds6.a = 6;
   f6(ds6); /* ds6 is copied into f6's actual parameter */

   return 0;
}
